/**
 *  Copyright (c) 2008-2011 Sonatype, Inc. All rights reserved.
 *
 *  This program is licensed to you under the Apache License Version 2.0,
 *  and you may not use this file except in compliance with the Apache License Version 2.0.
 *  You may obtain a copy of the Apache License Version 2.0 at http://www.apache.org/licenses/LICENSE-2.0.
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the Apache License Version 2.0 is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the Apache License Version 2.0 for the specific language governing permissions and limitations there under.
 */
package org.sonatype.jira.test;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import org.codehaus.plexus.util.IOUtil;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.InitializationError;
import org.junit.runners.model.Statement;
import org.sonatype.guice.bean.containers.Main;
import org.sonatype.inject.Sisu;
import org.sonatype.jira.internal.DefaultJiraXmlRpcProxy;
import org.sonatype.jira.mock.StubJira;
import org.sonatype.tests.http.runner.api.SuiteConfigurator;
import org.sonatype.tests.http.runner.junit.ConfiguratorMethod;

/**
 * Junit4 test runner using class annotations to determine the remote URL or replay file.
 *
 * @see Proxy
 * @see Record
 * @see Replay
 */
public class JiraProxyRunner
    extends BlockJUnit4ClassRunner
{
    private final List<SuiteConfigurator> configurators = new LinkedList<SuiteConfigurator>();

    private static boolean sisuInitialized = false;

    public JiraProxyRunner( final Class<?> klass )
        throws InitializationError
    {
        super( klass );
        if ( !JiraProxyTestCase.class.isAssignableFrom( klass ) )
        {
            IllegalArgumentException error =
                new IllegalArgumentException( "Can only run tests inheriting from JiraProxyTestCase." );
            throw new InitializationError( error );
        }
    }

    private <T extends Annotation> T getAnnotation( final Class<?> cls, final Class<T> annotation )
    {
        T a = cls.getAnnotation( annotation );
        if ( a != null && Arrays.asList( cls.getDeclaredAnnotations() ).contains( a ) )
        {
            return a;
        }
        else
        {
            return null;
        }
    }

    @Override
    protected List<FrameworkMethod> computeTestMethods()
    {
        // init sisu... there has to be some better way?
        if ( !sisuInitialized )
        {
            Main.main();
            sisuInitialized = true;
        }

        List<FrameworkMethod> methods = super.computeTestMethods();

        Class<?> testClass = getTestClass().getJavaClass();
        Record record = getAnnotation( testClass, Record.class );
        Proxy proxy = getAnnotation( testClass, Proxy.class );
        Replay replay = getAnnotation( testClass, Replay.class );
        Mock mock = getAnnotation( testClass, Mock.class );

        int numAnnos = 0;
        if ( record != null )
        {
            numAnnos += 1;
        }
        if ( proxy != null )
        {
            numAnnos += 1;
        }
        if ( replay != null )
        {
            numAnnos += 1;
        }
        if ( mock != null )
        {
            numAnnos += 1;
        }
        if ( numAnnos > 1 )
        {
            throw new IllegalArgumentException( "Only one annotation of Mock, Record, Proxy or Replay allowed." );
        }
        if ( numAnnos == 0 )
        {
            throw new IllegalArgumentException( "Exactly one annotation of Mock, Record, Proxy or Replay needed." );
        }

        List<FrameworkMethod> cfgMethods = new LinkedList<FrameworkMethod>();
        for ( FrameworkMethod method : methods )
        {
            DefaultJiraXmlRpcProxy jiraProxy = new DefaultJiraXmlRpcProxy();
            if ( replay != null )
            {
                String[] paths = replay.value();
                for ( String path : paths )
                {
                    String prefix = path;
                    prefix = prefix == null ? "" : prefix;
                    prefix += prefix.endsWith( "/" ) ? "" : "/";

                    cfgMethods.add( new ConfiguratorMethod( method.getMethod(),
                                                            new ReplayConfigurator( prefix, method.getMethod() ) ) );
                }
            }
            else if ( mock != null )
            {
                try
                {
                    StubJira stub = Sisu.lookup( StubJira.class );
                    if ( mock.value().length() > 0 )
                    {
                        String source = mock.value();
                        stub.setDatabase( IOUtil.toString( openStream( source ), "utf-8" ) );
                    }
                    cfgMethods.add( new ConfiguratorMethod( method.getMethod(), new MockConfigurator( stub ) ) );
                }
                catch ( Exception e )
                {
                    throw new IllegalArgumentException( e );
                }
            }
            else
            {
                String[] urls = record != null ? new String[] { record.value() } : proxy.value();
                for ( String url : urls )
                {
                    jiraProxy.setRecording( record != null );
                    cfgMethods.add( new ConfiguratorMethod( method.getMethod(),
                                                            new ProxyConfigurator( method.getMethod(), jiraProxy, url ) ) );
                }
            }
        }

        return cfgMethods;
    }

    private InputStream openStream( String source )
        throws FileNotFoundException
    {
        InputStream stream = getTestClass().getJavaClass().getResourceAsStream( source );
        if ( stream != null )
            return stream;

        stream = getTestClass().getJavaClass().getClassLoader().getResourceAsStream( source );
        if ( stream != null )
            return stream;

        return new FileInputStream( source );
    }

    @SuppressWarnings( "deprecation" )
    @Override
    protected void validateInstanceMethods( final List<Throwable> errors )
    {
        if ( computeTestMethods().isEmpty() && configurators != null && configurators.isEmpty() )
        {
            String msg = "No SuiteConfigurator found to run the tests with.";
            errors.add( new Exception( msg ) );
        }
        super.validateInstanceMethods( errors );
    }

    @Override
    protected Statement methodInvoker( final FrameworkMethod method, final Object test )
    {
        JiraProxyTestCase cfg = JiraProxyTestCase.class.cast( test );
        ConfiguratorMethod cfgMethod = ConfiguratorMethod.class.cast( method );
        JiraProxySuiteConfigurator configurator = JiraProxySuiteConfigurator.class.cast( cfgMethod.getConfigurator() );
        cfg.setConfigurator( configurator );
        cfg.handler(configurator.getHandler());

        return super.methodInvoker( method, test );
    }

}
